/************************************************
*     						*
*	KHELIDJ Ghalleb				*
*	Licence Informatique option MIME	*
*						*
*	Makefile				*
*						*
*	31 Mars 2000				*
*						*
************************************************/

%{
#include<string.h>

#define YYDEBUG 1 

typedef struct Fichier Fichier;
typedef struct Noeud Noeud;

void Etat(Noeud *);

Noeud* Recuperer(char *);
Noeud* Ajouter(Noeud*,Noeud*);

Noeud* Source(Noeud*,char*);
Noeud* Commande(Noeud*,char*);

char* CibleCat(char*,Noeud *);

%}

%union{

  int d;
  char * s;
  Noeud * n;
};

%token <s> FICH
%token  EQUI

%type <d> Chaine

%type <s> Action

%type <n> Cible
%type <n> Dependance
%type <n> Regle
%type <n> Equivalence


%left TAB
%left ' '
%left ':'

%right '='

%start Chaine
%%


/***********************************************
*              ANALYSEUR GRAMMATICAL           *
************************************************/


Chaine : /* rien */
               { $$ = 1; }
    | Chaine Regle '\n'
{ //printf("Regle %d\n",$$++);     /* Utile pour le deboguage, permet de connaitre le numero de regle */
  //Etat($2);                      /* Pour connaitre les caracteristiques des noeuds                  */  
}
    | Chaine '\n'
               { }
    ;

Equivalence : EQUI '(' FICH
               { $$ = Recuperer($3); }	           
            ;
	    
Cible : FICH '='
               { $1 = strcat($1,")");
		 $$ = Recuperer($1); }
      | FICH ':'
               { $$ = Recuperer($1); } 
      ;

Dependance : Cible FICH 
               { $$ = Source($1,$2); }
             | Dependance FICH
	       { $$ = Source($1,$2); }
             | Cible Equivalence
               { $$ = Ajouter($1,$2); }
             | Dependance Equivalence 
               { $$ = Ajouter($1,$2); }
             ;

Action : TAB FICH
               { $$ = $2; }
            | Action FICH
               { $$ = (char *)strcat($1," ");
	         $$ = (char *)strcat($1,$2); }
            | TAB Equivalence
               { $$ = CibleCat("",$2); }
            | Action Equivalence
               { $$ = (char *)CibleCat($1,$2); }
            ;

Regle : Dependance '\n' Action 
               { $$ = Commande($1,$3); }
       | Regle '\n' Action
               { $$ = Commande($1,$3); }
       ;


%%


/************************************************ 
*             FONCTIONS PRINCIPALES             *
*************************************************/

#include <stdio.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>

static int ligne = 1;
FILE * makefile;

struct Fichier* tableFichiers[50];

int nbFichiers=0;

void execution (char *,int);
char* recup_output(char *);


int 
main(int argc,char* argv[]){

  char* file_make;
  char* cible;
  int option;

  file_make = "makefile";
  cible = NULL;


switch(argc){

  case 1:
    option = 0;
    break;

  case 2:    
    if(!strcmp(argv[1],"clean")){
      system("rm *.o");
      return 0;
    }
    else if(!strcmp(argv[1],"-s")){
      option = 1;
      break;
    }
    else if(!strcmp(argv[1],"-n")){
      option = 2;
      break;
    }
    else if(!strcmp(argv[1],"-f")){
      file_make = argv[2];
      cible = NULL;
      option = 0;
      break;
    }
     else {
      cible = argv[1];
      option = 0;
      break;}

  case 3: 
    if(!strcmp(argv[1],"-s")){
      option = 1;
      cible=argv[2];
      break;
    }
    else if(!strcmp(argv[1],"-n")){
      option = 2;
      cible=argv[2];
      break;
    }
    else if(!strcmp(argv[1],"-f")){
      file_make = argv[2];
      cible = NULL;
      option = 0;
      break;
    }
    else{
     printf("%s [-s] [-n] [clean] [-f nom_du_makefile] [cible]" ,argv[0]);
     exit(1);
    }
     
  case 4:
    if(!strcmp(argv[1],"-f")){
      file_make = argv[2];
      cible = argv[3];
      option =  0;
      break;
    }
    else{
     printf("%s [-s] [-n] [clean] [-f nom_du_makefile] [cible]" ,argv[0]);
     exit(1);
    }

  default:
     printf("%s [-s] [-n] [clean] [-f nom_du_makefile] [cible]" ,argv[0]);
     exit(1);
  }


  
  makefile = fopen(file_make,"r");


  assert(makefile!=0);
  yyparse();
  fclose(makefile);
  execution(cible,option); 
  return 0;

}


/************************************************** 
*               ANALYSEUR LEXIQUAL                *
***************************************************/

int 
yyerror(char *s){

  fprintf(stderr, "%d: %s\n",ligne,s);
  return 0;
}


int 
yylex(){

  int c;
  char *Chaine;
  char tab[256];
  
re:
  switch(c = fgetc(makefile)){

  case '#':                      
    while((c = fgetc(makefile)) != ('\n'));
    goto re;

  case '(':
    return '(';

  case ' ':
    goto re;

  case '\t':
    return TAB;
    
  case ':':
    return ':';
    
  case '$':
    return EQUI;

  case '=':
    return '=';
  
  case '\n':
    return '\n';
    
  case EOF:
    return 0;
    
  default:
    ungetc(c,makefile);
    assert(fscanf(makefile,"%s", tab)==1);
    Chaine = strdup(tab);
    yylval.s = Chaine;
    return FICH;
  }
}


/***********************************************************
*           CONSTRUCTION DE L'ARBRE SYNTAXIQUE             *
************************************************************/

/* *** Structure des Fichiers *** */

struct Fichier{

  char* nom;
  Noeud* construction;
};


/* *** Structure des Feuilles *** */

struct Noeud{

  Fichier* cible;
  int nbFichiersSources;
  Fichier* FichiersSources[10];
  int nbCommandes;
  char* Commandes[1];

};

/* *** Recuperer le Fichier *** */

Fichier* 
recupererFichier(char* nom){

  int i;
  Fichier *f=NULL;

  for(i=0;i<nbFichiers;i++){
    if(strcmp(nom,tableFichiers[i]->nom)==0){
      f = tableFichiers[i];
      break;
    }
  }
  if(f==NULL){
    
    tableFichiers[nbFichiers] = (Fichier*)malloc(sizeof(Fichier));    
    tableFichiers[nbFichiers]->nom = nom;
    tableFichiers[nbFichiers]->construction = NULL;
    f = tableFichiers[nbFichiers];
    nbFichiers++;
  }
  return f;
}

/* ***  Recuperer la Cible  *** */

Noeud* 
Recuperer(char* nom){

  Noeud * n;
  Fichier* f=recupererFichier(nom);

  if(f->construction == NULL){ 
    
    f->construction = (Noeud*)malloc(sizeof f->construction[0]);
    assert(f->construction!=0);
    n = f->construction;
    n->cible = f;
    n->nbFichiersSources = 0;
    
  }
  else{

    n = f->construction;
  }
  return n;
}

/* *** Ajouter un Source *** */

Noeud* 
Source(Noeud* n,char* nom){

  n->FichiersSources[n->nbFichiersSources] = recupererFichier(nom);
  n->nbFichiersSources++;
  n->FichiersSources[n->nbFichiersSources]=(Fichier*)malloc(sizeof(Fichier*));
  assert(n->FichiersSources[n->nbFichiersSources] != 0);
  return n;
}

/* *** Ajouter une Commande *** */

Noeud* 
Commande(Noeud* n, char* com){

  n->Commandes[n->nbCommandes] = (char*)malloc(sizeof (com));
  assert(n->Commandes[n->nbCommandes] != 0);
  n->Commandes[n->nbCommandes] = com;
  n->nbCommandes++;
  return n;
}


/* *** Ajouter une Cible *** */

Noeud*
Ajouter(Noeud* n1,Noeud* n2){

  int i;
  char * nom;
  printf("nombre fichiers sources :%d\n",n2->nbFichiersSources);
  for(i=0;i<(n2->nbFichiersSources);i++){
    nom = n2->FichiersSources[i]->nom;


    n1 = Source(n1,nom);
  }
  return n1;
}


/* *** Afficher les Noeuds *** */

void
Etat(Noeud* n){

  int i;
  printf("nom:%s\n",n->cible->nom);
  printf("nombre fichiers sources:%d\n",n->nbFichiersSources);
  for(i=0;i<n->nbFichiersSources;i++){
    printf("fichier %d:%s\n",i+1,n->FichiersSources[i]->nom);
  }
  for(i=0;i<(n->nbCommandes);i++){
    printf("commande %d:%s\n",i+1,n->Commandes[i]);
  }
  printf("fin\n\n\n");

}




/* *** Concatener des Sources avec une Chaine *** */
 
char *
CibleCat(char* n1,Noeud* n2){

  int i;
  char* cat;
  cat = (char*)strcat(n1," ");
  cat = (char*)strcat(cat,n2->FichiersSources[0]->nom);
  for(i=1;i<n2->nbFichiersSources;i++){
    cat = strcat(cat," ");
    cat = strcat(cat,n2->FichiersSources[i]->nom);
  }

  return cat;


}



/* *** Executer  *** */

void 
execution(char* output,int option){

  int i,j;
  Fichier * f;
  Noeud * cible;
  struct stat temps;
  time_t *time3;
  time_t time1;
  time_t time2;
  int time;
  
  if(output == NULL)   
    output = recup_output(output);

  //  printf("option:%d\n",option);                   /* pour savoir quel option est choisie */

  cible = Recuperer(output);
   
  for(i=0;i<cible->nbFichiersSources;i++){
    f=cible->FichiersSources[i];
     if(f->construction != NULL)
      execution(f->nom,option);
  }

  stat(cible->cible->nom,&temps);
  time1 = temps.st_mtime;                  

  for(i=0;i<cible->nbFichiersSources;i++){   

    stat(cible->FichiersSources[i]->nom,&temps);
    time2 = temps.st_mtime;                

    if((time2) > (time1)){                 
      for(j=0;j<cible->nbCommandes;j++){
	if(option!=1)
	  printf("------>   %s\n",cible->Commandes[j]);
	if(option!=2)
	  system(cible->Commandes[j]);      
      }                                    
      break;
    }
  }
}





/* *** Recuperer la Sortie *** */

char*
recup_output(char *output){
  int i=0;
 
  while(i<nbFichiers){      
                                              
    if(tableFichiers[i]->construction != NULL)
      if(tableFichiers[i]->construction->Commandes[0] != NULL)
	return (tableFichiers[i]->nom);     
    i++;
  }
  
  if(output == NULL){               
    printf("Le Makefile a un probleme\n");
    exit(1);
  }
}










